from abc import ABCMeta, abstractmethod
import re
import sys
import typing as t
from logging import getLogger
from collections import ChainMap, abc
from typing_extensions import Self
from weakref import WeakKeyDictionary
from importlib import import_module
from .exceptions import ProError


from . import signals
from .markers import (
    GUARDED,
    PRIVATE,
    PROTECTED,
    PUBLIC,
    AccessModifier,
    _PredicateOpsMixin,
    Injectable,
    ProPredicate,
    is_injectable,
)
from ._common import ReadonlyDict, ordered_set, private_setattr, FrozenDict
from .providers import Provider, ProviderRegistryMixin

from .graph.core import Graph, DepKey, DepSrc


logger = getLogger(__name__)

_dict_setdefault = dict.setdefault
_dict_setitem = dict.__setitem__
_object_new = object.__new__


def _calling_module(depth=2) -> t.Optional[str]:
    """Get the globals() or locals() scope of the calling scope"""
    name = sys._getframe(depth).f_globals.get("__name__")
    try:
        name and import_module(name)
    except Exception:  # pragma: no cover
        return
    else:
        if name == __name__:
            name = _calling_module(depth + 2)
        return name


class _ContainerRegistry(ReadonlyDict[str, dict["BaseContainer", None]]):
    __slots__ = ()

    __getitem = dict[str, dict["BaseContainer", None]].__getitem__
    __get = dict[str, dict["BaseContainer", None]].get
    __contains = dict[str, dict["BaseContainer", None]].__contains__

    _placeholders = {
        "**": r".*",  # '[^{placeholders}]*',
        "*": r"[^.|:]*",  #'[^{placeholders}.]*',
        "++": r".+",  # '[^{placeholders}]+',
        "+": r"[^.|:]+",  # '[^{placeholders}.]+',
    }

    _pattern_split_re = re.compile(f"({'|'.join(map(re.escape, _placeholders))})")
    pfmt = {"placeholders": re.escape("".join({*"".join(_placeholders)}))}
    for k in _placeholders:
        _placeholders[k] = _placeholders[k].format_map(pfmt)
    del k, pfmt

    @classmethod
    def _complie_pattern(cls, pattern: t.Union[str, re.Pattern]):
        if isinstance(pattern, re.Pattern):
            return pattern
        parts = cls._pattern_split_re.split(pattern)
        pattern = "".join(cls._placeholders.get(x, re.escape(x)) for x in parts)
        return re.compile(f"^{pattern}$")

    def add(self, *instances: "BaseContainer"):
        for inst in instances:
            inst._is_anonymous or self.__getitem(inst.qualname).setdefault(inst)

    def get(self, key: str, default=None):
        ret = self.__get(key)
        return default if ret is None else tuple(ret)

    @t.overload
    def all(
        self,
        *patterns: t.Union[str, re.Pattern],
        module: t.Union[str, list[str], tuple[str]] = None,
        name: t.Union[str, list[str], tuple[str]] = None,
        group: t.Literal[False] = False,
    ) -> abc.Generator["BaseContainer", None, None]:
        ...  # pragma: no cover

    @t.overload
    def all(
        self,
        *patterns: t.Union[str, re.Pattern],
        module: t.Union[str, list[str], tuple[str]] = None,
        name: t.Union[str, list[str], tuple[str]] = None,
        group: t.Literal[True] = True,
    ) -> abc.Generator[tuple["BaseContainer"], None, None]:
        ...  # pragma: no cover

    def all(
        self,
        *patterns: t.Union[str, re.Pattern],
        module: t.Union[str, list[str], tuple[str]] = None,
        name: t.Union[str, list[str], tuple[str]] = None,
        group: bool = False,
    ):

        if module or name:
            if not isinstance(module, (list, tuple)):
                module = (f'{module or "**"}',)
            if not isinstance(name, (list, tuple)):
                name = (f'{name or "**"}',)

            lm, ln = len(module), len(name)
            if lm == ln:
                it = zip(module, name)
            elif lm == 1:
                it = zip(module * ln, name)
            elif ln == 1:
                it = zip(module, name * lm)
            else:
                raise ValueError(
                    f"`module` and `name` lists can either be of equal lengths "
                    f"or one of them must contain a single item."
                )
            patterns = tuple(":".join(mn) for mn in it) + patterns
        elif not patterns:
            patterns = ("**",)

        seen = set()
        for pattern in (self._complie_pattern(p) for p in patterns):
            for k, v in self.items():
                if not (k in seen or not pattern.search(k) or seen.add(k)):
                    if group:
                        yield tuple(v)
                    else:
                        yield from v

    @t.overload
    def find(
        self,
        *patterns: t.Union[str, re.Pattern],
        module: str = None,
        name: str = None,
        group: bool = True,
    ):
        ...  # pragma: no cover

    def find(self, *a, **kw):
        for v in self.all(*a, **kw):
            return v

    def __contains__(self, k: t.Union[str, "BaseContainer"]):
        if isinstance(k, BaseContainer):
            return k in self.__get(k.qualname, ())
        return self.__contains(k)

    def __getitem__(self, k: str) -> tuple["BaseContainer"]:
        return tuple(self.__getitem(k))

    def __missing__(self, key: str):
        return _dict_setdefault(self, key, WeakKeyDictionary())

    def __repr__(self):
        return f"{self.__class__.__name__}({({k: self[k] for k in self})})"


class ProEntrySet(FrozenDict["BaseContainer", None]):

    __slots__ = ()

    __contains = dict.__contains__
    is_atomic: bool = False

    @classmethod
    def make(cls, it: abc.Iterable["BaseContainer"] = ()):
        return it if isinstance(it, cls) else cls((v, None) for v in it)

    fromkeys = make

    def atomic(self):
        return AtomicProEntrySet.make(self)

    def _eval_hashable(self):
        return tuple(self.atomic())

    __hash__ = FrozenDict.__hash__

    def __contains__(self, k: "BaseContainer") -> bool:
        if self.__contains(k):
            return True
        elif getattr(k, "is_atomic", True):
            return not self.is_atomic and any(k in c for c in self if not c.is_atomic)
        else:
            return all(a in self for a in k.atomic)

    def __eq__(self, o: Self) -> bool:
        if isinstance(o, ProEntrySet):
            return self._eval_hashable() == o._eval_hashable()
        elif isinstance(o, abc.Mapping):
            return False
        return NotImplemented

    def __ne__(self, o: Self) -> bool:
        if isinstance(o, ProEntrySet):
            return self._eval_hashable() != o._eval_hashable()
        elif isinstance(o, abc.Mapping):
            return True
        return NotImplemented


class AtomicProEntrySet(ProEntrySet):

    __slots__ = ()

    is_atomic: bool = True

    @classmethod
    def make(cls, it: abc.Iterable["Container"] = ()):
        return (
            it
            if isinstance(it, cls)
            else cls((a, None) for at in it for a in at.atomic)
        )

    def atomic(self):
        return self

    def _eval_hashable(self):
        return tuple(self)


class ContainerMeta(ABCMeta):

    _registry: t.ClassVar[_ContainerRegistry] = _ContainerRegistry()
    register: t.Final = ABCMeta.register

    def __call__(self, *args, **kwds):
        if not "module" in kwds:
            kwds["module"] = _calling_module()
        res: Container = super().__call__(*args, **kwds)
        self._registry.add(res)
        return res


@ProPredicate.register
@private_setattr(frozen="_pro")
class BaseContainer(_PredicateOpsMixin, metaclass=ContainerMeta):

    __slots__ = ()
    is_atomic: bool = True

    @classmethod
    @abstractmethod
    def _collect(cls, *a, **kw) -> "Group":
        ...  # pragma: no cover

    @property
    def _is_anonymous(self) -> bool:
        """`True` if this container can be added to the registry or `False` if
        otherwise.
        """
        return not self.name

    @property
    @abstractmethod
    def atomic(self):
        """`AtomicProEntries`(s)"""

    @property
    @abstractmethod
    def bases(self) -> abc.Mapping[Injectable, Provider]:
        """the base container."""

    @property
    @abstractmethod
    def providers(self) -> abc.Mapping[Injectable, Provider]:
        """A mapping of providers registered in the container."""

    @property
    @abstractmethod
    def _pro(self) -> t.Optional[FrozenDict[Self, int]]:
        """The container's provider resolution order."""

    @property
    def pro(self) -> FrozenDict[Self, int]:
        """The container's provider resolution order.

        Like python's class `__mro__` the `pro` is computed using
        [C3 linearization](https://en.wikipedia.org/wiki/C3_linearization)

        Returns:
            pro (FrozenDict[Container, int]):
        """
        if not None is (pro := self._pro):
            return pro
        self.__setattr(_pro=self._evaluate_pro())
        return self._pro

    @property
    def qualname(self) -> str:
        return f"{self.module}:{self.name}"

    @property
    @abstractmethod
    def g(self) -> ReadonlyDict["Graph", "Graph"]:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def module(self) -> str:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def name(self) -> str:
        ...  # pragma: no cover

    def extends(self, other: Self) -> bool:
        """Check whether this container extends the given base.

        Params:
            base (Container): The base container to check

        Returns:
            bool:
        """
        return other in self.pro
        # if other.is_atomic:
        #     return other in self.pro
        # else:
        #     return all(self.extends(x) for x in other.atomic)

    def get_graph(self, base: "Graph"):
        try:
            return self.g[base]
        except KeyError:
            return _dict_setdefault(self.g, base, self.create_graph(base))

    def create_graph(self, base: "Graph"):
        return Graph(self, base)

    def pro_entries(
        self, it: abc.Iterable["Container"], graph: "Graph", src: "DepSrc"
    ) -> abc.Iterable["Container"]:
        pro = self.pro
        return tuple(c for c in it if c in pro)

    def _evaluate_pro(self):
        if self.is_atomic:
            res, bases = {self: 0}, [*self.bases]
        else:
            res, bases = {}, [*self.atomic]

        if bases:
            i, miss = 0, 0
            ml = [*([*b.pro] for b in bases), [*bases]]
            while ml:
                if i == len(ml):
                    if miss >= i:
                        raise ProError(
                            f"Cannot create a consistent provider resolution order {miss=}, {ml=}"
                        )
                    i = 0
                ls = ml[i]
                h = ls[0]
                if h in res:
                    pass
                elif any(l.index(h) > 0 for l in ml if not l is ls and h in l):
                    i += 1
                    miss += 1
                    continue
                else:
                    res[h] = i
                ls.pop(0)
                miss = 0
                if ls:
                    i += 1
                else:
                    ml.pop(i)

        return AtomicProEntrySet.fromkeys(res)

    def __eq__(self, o) -> bool:
        if isinstance(o, BaseContainer):
            return self is o
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, BaseContainer):
            return not self is o
        return NotImplemented

    def __hash__(self) -> int:
        return id(self)

    def __or__(self, o):
        if isinstance(o, BaseContainer):
            return self._collect((self, o))
        else:
            return super().__or__(o)

    __ior__ = __or__

    # def __ror__(self, o):
    #     if isinstance(o, BaseContainer):
    #         return self._collect((o, self))
    #     else:
    #         return super().__ror__(o)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.qualname!r})"


class Container(BaseContainer, ProviderRegistryMixin):
    """A mapping of dependencies to their providers. We use them to bind
    dependencies to their providers.

    Attributes:
        name (str): The container's name
        bases (tuple[Container]): The container's bases
        default_access_modifier (AccessModifier): The default `access_modifier` to assign
        to providers registered in this container
    """

    __slots__ = (
        "module",
        "name",
        "providers",
        "bases",
        "default_access_modifier",
        "g",
        "_pro",
        "__weakref__",
    )

    name: str
    bases: ProEntrySet
    default_access_modifier: AccessModifier
    g: ReadonlyDict["Graph", "Graph"]
    providers: ReadonlyDict[Injectable, Provider]
    _pro: FrozenDict[Self, int]
    is_atomic: t.Final = True

    def __init__(
        self,
        name: str = None,
        *bases: Self,
        module: str,
        access_modifier: AccessModifier = PUBLIC,
    ) -> None:
        """Create a container.

        Params:
            name (str, optional): Name of the container
            *bases (Container, optional): Base container.
            access_modifier (AccessModifier, optional): The default `access_modifier`
                to assign providers
        """
        if name and not name.isidentifier():
            raise ValueError(f"name must be a valid identifier not {name!r}")

        self.__setattr(
            _pro=None,
            bases=ProEntrySet(),
            name=name or f"__anonymous__",
            providers=ReadonlyDict(),
            module=module,
            g=ReadonlyDict(),
            default_access_modifier=AccessModifier(access_modifier),
        )

        bases and self.extend(*bases)
        signals.on_container_create.send(self.__class__, container=self)

    @property
    def atomic(self):
        """`AtomicProEntries`(s)"""
        return AtomicProEntrySet(((self, None),))

    def extend(self, *bases: Self) -> Self:
        """Adds containers to extended by this container.
        Args:
            *bases (Container): The base containers to be extended

        Returns:
            Self: this container
        """
        self.__setattr(bases=self.bases | ProEntrySet.make(bases))
        return self

    def access_modifier(self, accessor: Self):
        """Get the `AccessModifier`

        Params:
            accessor (Container):

        Returns:
            access_modifier (AccessModifier):
        """
        if accessor is self:
            return PRIVATE
        elif self.extends(accessor):
            return GUARDED
        elif accessor.extends(self):
            return PROTECTED
        else:
            return PUBLIC

    @classmethod
    def _collect(cls, *a, **kw) -> "Group":
        return Group(*a, **kw)

    def _on_register(self, abstract: Injectable, provider: Provider):
        pass

    def __contains__(self, x):
        return x in self.providers or any(x in b for b in self.bases)

    def __setitem__(self, key: Injectable, provider: Provider) -> Self:
        """Register a dependency provider

            container[_T] = providers.Value('abc')

        Params:
            abstract (Injectable): The dependency to be provided
            provider (Provider): The provider to provide the dependency
        """
        if not is_injectable(key):
            raise TypeError(
                f"expected `Injectable` not. `{key.__class__.__qualname__}`"
            )

        if prov := provider._setup(self, key):
            self._on_register(key, prov)
            _dict_setitem(self.providers, key, prov)
            # self.__setitem(key, prov)
            signals.on_provider_registered.send(self, abstract=key, provider=provider)

    def __getitem__(self, k):
        try:
            return self.providers[k]
        except KeyError:
            if isinstance(k, Provider) and (k.container or self) is self:
                return k

    def _resolve(self, key: "DepKey", graph: "Graph"):
        if prov := self[key.abstract]:
            access = prov.access_modifier or self.default_access_modifier

            if access in self.access_modifier(key.container):
                if prov._can_resolve(key, graph):
                    return (prov,)
        return ()

    def __bool__(self):
        return True


class Group(BaseContainer):
    """A `Container` group."""

    __slots__ = (
        "_g",
        "bases",
        "name",
        "module",
        "_pro",
        "_is_anonymous",
        "__weakref__",
    )
    is_atomic = False

    bases: ProEntrySet
    _pro: FrozenDict[Self, int]

    def __new__(
        cls: type[Self],
        bases: t.Union[abc.Iterable[Container], Self] = (),
        *,
        name: str = None,
        module: str,
    ) -> Self:
        self = _object_new(cls)
        # typ: type[bases] = bases.__class__
        # if issubclass(typ, cls):
        #     if not name and module == bases.module:
        #         return bases
        #     else:
        #         bases = bases.bases
        # else:
        #     bases = ProEntrySet.make(bases)

        self.__setattr(
            _pro=None,
            bases=ProEntrySet.make(bases),
            _is_anonymous=not name,
            module=module,
            name=name or f'[{"|".join(ordered_set(c.qualname for c in bases))}]',
        )
        return self

    @property
    def g(self):
        try:
            return self._g
        except AttributeError:
            self.__setattr(_g=ReadonlyDict())
            return self._g

    @property
    def atomic(self):
        return self.bases.atomic()

    @property
    def qualname(self) -> None:
        return f"{self.module}:{self.name}"

    @property
    def providers(self):
        return ChainMap(*(a.providers for a in self.bases))

    @classmethod
    def _collect(cls, *a, **kw):
        return cls(*a, **kw)

    def __sub__(self, o):
        if isinstance(o, BaseContainer):
            check = o.atomic
            return self._collect(x for x in self.atomic if not x in check)
        return NotImplemented

    __isub__ = __sub__

    # def __rsub__(self, o):
    #     if isinstance(o, Group):
    #         check = self.atomic
    #         return self._collect(x for x in o.atomic if not x in check)
    #     return NotImplemented

    def __bool__(self):
        return not not self.bases

    def __contains__(self, x):
        return any(x in b for b in self.bases)


ContainerMeta.register = None



from contextvars import ContextVar
from logging import getLogger
from threading import Lock, local
import typing as t

from typing_extensions import Self


from ._common import private_setattr

from .exceptions import InvalidStateError, InvalidStateError
from .containers import BaseContainer, Container
from .graph.core import Graph, _null_graph
from .injectors import Injector, NullInjector, _null_injector

logger = getLogger(__name__)


_T_Injector = t.TypeVar("_T_Injector", bound=Injector)
_T_Initial = t.Union[_T_Injector, t.Literal[_null_injector]]  # type: ignore


@private_setattr(frozen="is_active")
class Scope(t.Generic[_T_Injector]):
    """An isolated dependency resolution `scope` for a given container.

    Scopes assemble the dependency graphs of dependencies registered in their container.

    Attributes:
        container (Container): The container who's scope we are creating
        parent (Scope): The parent scope. Defaults to None

    Args:
        container (Container): The container who's scope we are creating
        parent (Scope, optional): The parent scope. Defaults to NullScope

    """

    __slots__ = (
        "graph",
        "current",
        "initial",
        "parent",
    )

    graph: Graph
    parent: Self
    current: _T_Injector
    initial: _T_Initial[_T_Injector]

    _injector_class: type[_T_Injector] = Injector

    def __init__(
        self,
        graph: t.Union[Container, Graph],
        parent: Self = None,
        *,
        initial: _T_Initial[_T_Injector] = None,
        **kwargs,
    ) -> None:
        parent = parent or _null_scope

        if isinstance(graph, BaseContainer):
            graph = graph.get_graph(parent.graph)
        elif isinstance(graph, Graph):
            if not graph.parent is parent.graph:
                raise ValueError(f"graph mismatch")
        else:
            raise TypeError(
                f"first argument must be a `Container`, `DepGraph` not `{graph.__class__.__name__}` "
            )

        if initial is None:
            initial = _null_injector

        attrs = (
            self.__default_attrs__()
            | kwargs
            | {"parent": parent, "graph": graph, "initial": initial}
        )

        self.__init_attrs__(attrs)
        self._set_current(initial)

    @property
    def container(self):
        return self.graph.container

    @property
    def name(self):
        return self.container.qualname

    @property
    def active(self):
        return not self.initial is self.current

    def __getitem__(self, key):
        return self.graph[key]

    def __init_attrs__(self, kwds: dict):
        self.__setattr(**kwds)

    def __default_attrs__(self):
        return {}

    def injector(self, *, push=True) -> _T_Injector:
        if inj := self.current:
            return inj
        elif push:
            return self.push()
        else:
            return self._new_injector()

    def _new_injector(self):
        return self._injector_class(self.graph, self.parent.injector())

    def push(self):
        if self.active:
            raise InvalidStateError(f"injector already running: {self}")
        return self._push(force=True)

    def _push(self, *, force: bool = False):
        if force or not self.active:
            inj = self._new_injector()
            self._set_current(inj)
            return inj
        return self.current

    def pop(self):
        if not self.active:
            raise InvalidStateError(f"injector not running: {self}")
        return self._pop(force=True)

    def _pop(self, *, force: bool = False):
        if force or self.active:
            self.current.close()
            self._set_current(self.initial)

    def _set_current(self, injector: _T_Injector):
        self.__setattr("current", injector, injector is self.initial)

    def __eq__(self, o) -> bool:
        if isinstance(o, Scope):
            return o is self
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, Scope):
            return not o is self
        return NotImplemented

    def __hash__(self):
        return id(self)

    def __str__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name!r}, parent="{self.parent!s}")'

    def __enter__(self):
        return self.injector()

    def __exit__(self, *err):
        self.pop()
        return err and err[0] != None or False


class ThreadSafeScope(Scope[_T_Injector]):
    """A thread safe `Scope` implementation"""

    __slots__ = ("lock",)

    lock: Lock

    def __init_attrs__(self, kwds: dict):
        kwds["lock"] = Lock()
        return super().__init_attrs__(kwds)

    def push(self) -> _T_Injector:
        if self.active:
            raise InvalidStateError(f"injector already running: {self}")

        with self.lock:
            return self._push()

    def pop(self):
        if self.active:
            with self.lock:
                return self._pop()

        raise InvalidStateError(f"injector not running: {self}")


class _ThreadLocalState(local, t.Generic[_T_Injector]):

    __slots__ = ("__initial",)

    injector: _T_Injector

    def __init__(self, injector=_null_injector) -> None:
        self.injector = injector


class ThreadLocalScope(Scope[_T_Injector]):
    """A scope that uses `threading.local` to manage injectors"""

    __slots__ = ("__local",)

    __local: _ThreadLocalState[_T_Injector]

    @property
    def current(self):
        return self.__local.injector

    def __init_attrs__(self, kwds):
        self.__local = _ThreadLocalState(kwds["initial"])
        return super().__init_attrs__(kwds)

    def _set_current(self, injector: _T_Injector) -> _T_Injector:
        self.__local.injector = injector
        return injector


class _NullContextVar:
    __slots__ = ()

    def get(self):
        return _null_injector


_null_context_var = _NullContextVar()


class ContextLocalScope(Scope[_T_Injector]):
    """A scope that uses `contextvars.ContextVar` to manage injectors"""

    __slots__ = ("__var",)

    __var: ContextVar[_T_Injector]

    def __init_attrs__(self, kwds):
        self.__var = _null_context_var
        super().__init_attrs__(kwds)
        self.__var = ContextVar(f"{self.name}.injector", default=self.initial)

    @property
    def current(self):
        return self.__var.get()

    def _set_current(self, injector: _T_Injector) -> _T_Injector:
        self.__var.set(injector)


class NullScope(Scope[NullInjector]):
    """A 'noop' `Scope` used as the parent of root scopes.

    Attributes:
        parent (None): The parent scope
        graph (NullGraph):
    """

    __slots__ = ()
    parent = None
    level = -1
    graph = _null_graph
    name = "<null>"

    def __init__(self) -> None:
        ...  # pragma: no cover

    def __bool__(self):
        return False

    def __eq__(self, o) -> bool:
        return o.__class__ is self.__class__

    def __ne__(self, o) -> bool:
        return not o.__class__ is self.__class__

    __hash__ = classmethod(hash)

    def injector(self, *, setup=True, create=True):
        return _null_injector


_null_scope = NullScope()



from ._common import Missing


from .markers import (
    Injectable,
    is_injectable,
    T_Injected,
    T_Injectable,
    Dep,
    DependencyMarker,
    Lookup,
    PureDep,
)


from .exceptions import InjectorLookupError, UziException


from . import injectors, providers
from .containers import Container
from .injectors import Injector
from .providers import Provider
from .scopes import Scope, ThreadSafeScope, ThreadLocalScope, ContextLocalScope


from . import _receivers



import typing as t
from abc import abstractmethod
from collections.abc import Callable, Hashable

from . import FrozenDict, private_setattr


_T_Obj = t.TypeVar("_T_Obj")
_T_Expr = t.TypeVar("_T_Expr")
_T_Default = t.TypeVar("_T_Default")
_T_Args = t.TypeVar("_T_Args")
_T_Kwargs = t.TypeVar("_T_Kwargs")
_T_Item = t.TypeVar("_T_Item", int, str, Hashable)

_object_new = object.__new__


class ExpressionError(Exception):
    ...  # pragma: no cover


class EvaluationError(ExpressionError):
    @classmethod
    def wrap(cls, exc: Exception):
        if _cls := _eval_exc_types.get(exc.__class__):
            return exc if _cls is exc.__class__ else _cls(exc)
        elif cls is EvaluationError:
            for e in _eval_exc_types:
                if isinstance(exc, e):
                    return _eval_exc_types[e](exc)
                    break
            return exc  # (_cls or cls)(exc)
        else:
            return cls(exc)


class AttributeEvaluationError(EvaluationError, AttributeError):
    ...  # pragma: no cover


class KeyEvaluationError(EvaluationError, KeyError):
    ...  # pragma: no cover


class IndexEvaluationError(EvaluationError, IndexError):
    ...  # pragma: no cover


class CallEvaluationError(EvaluationError, TypeError):
    ...  # pragma: no cover


_eval_exc_types = {
    AttributeError: AttributeEvaluationError,
    KeyError: KeyEvaluationError,
    IndexError: IndexEvaluationError,
    LookupError: EvaluationError,
    CallEvaluationError: CallEvaluationError,
}


@private_setattr
class Expression(t.Generic[_T_Expr, _T_Obj]):
    __slots__ = ("__expr__",)

    __expr__: _T_Expr

    __evaluation_errors__ = (EvaluationError,)

    def __new__(cls, expr: _T_Expr):
        self = _object_new(cls)
        self.__setattr(__expr__=expr)
        return self

    def __eq__(self, x):
        if isinstance(x, self.__class__):
            return x.__expr__ == self.__expr__
        # elif isinstance(x, self.__expr__.__class__):
        #     return x == self.__expr__
        return False

    def __ne__(self, x):
        return not self == x

    def __repr__(self):
        return f"{self.__class__.__name__}({self!s})"

    def __str__(self):
        return str(self.__expr__)

    def __hash__(self):
        return hash(self.__expr__)

    def __reduce__(self):
        return self.__class__, (self.__expr__,)

    @abstractmethod
    def __eval__(self, o: _T_Obj):  # pragma: no cover
        raise NotImplementedError(f"{self.__class__.__name__}.__eval__datapath__()")

    # def __setattr__(self, name: str, value) -> None:
    #     if hasattr(self, "__expr__"):
    #         getattr(self, name)
    #         raise AttributeError(f"cannot set readonly attribute {name!r}")

    #     return super().__setattr__(name, value)


class Attribute(Expression[str, _T_Obj]):
    __slots__ = ()

    __evaluation_errors__ = (AttributeError,)

    def __eval__(self, o: _T_Obj):
        __tracebackhide__ = True
        return getattr(o, self.__expr__)

    def __str__(self):
        return f".{self.__expr__}"


class Item(Expression[_T_Item, _T_Obj]):
    __slots__ = ()

    __evaluation_errors__ = KeyError, IndexError

    def __eval__(self, o: _T_Obj):
        __tracebackhide__ = True
        return o[self.__expr__]

    def __str__(self):
        return f"[{self.__expr__!r}]"


class Slice(
    Expression[
        tuple[t.Union[_T_Item, None], t.Union[_T_Item, None], t.Union[_T_Item, None]],
        _T_Obj,
    ]
):

    __slots__ = ()

    __evaluation_errors__ = KeyError, IndexError

    def __eval__(self, o: _T_Obj):
        __tracebackhide__ = True
        return o[slice(*self.__expr__)]

    def __str__(self):
        start, stop, step = ("" if v is None else f"{v!r}" for v in self.__expr__)
        return f"[{start}:{stop}:{step}]"


class Call(Expression[tuple[_T_Args, _T_Kwargs], _T_Obj]):
    __slots__ = ()

    __evaluation_errors__ = (CallEvaluationError,)

    def __eval__(self, o: _T_Obj):
        __tracebackhide__ = True
        args, kwargs = self.__expr__
        try:
            return o(*args, **kwargs)
        except TypeError as e:
            if isinstance(o, Callable):
                raise

            raise CallEvaluationError(e) from e

    def __str__(self):
        args, kwargs = self.__expr__
        a = ", ".join(map(repr, args))
        kw = ", ".join(f"{k!s}={v!r}" for k, v in kwargs.items())
        return f'({", ".join(filter(None, (a, kw)))})'


class Lookup(Expression[tuple[Expression[t.Any, _T_Obj]], _T_Obj], t.Generic[_T_Obj]):
    """A chain of lookup experesions."""

    __slots__ = ("__expr__",)

    __offset__ = None

    __expr__: tuple[Expression[_T_Expr, _T_Obj]]

    def __new__(cls, *ops: Expression[_T_Expr, _T_Obj]):
        self = _object_new(cls)
        self.__setattr(__expr__=ops)
        return self

    @property
    def __ops__(self) -> None:
        return self.__expr__[self.__offset__ :]

    def __push__(self, *expr: Expression[_T_Expr, _T_Obj]):
        return self.__class__(*self.__expr__, *expr)

    def __getattr__(self, name: str):
        if name[:2] == "__" == name[-2:]:
            raise AttributeError(name)

        return self.__push__(Attribute(name))

    def __getitem__(self, k: t.Union[_T_Item, slice]):
        if k.__class__ is slice:
            return self.__push__(Slice((k.start, k.stop, k.step)))
        else:
            return self.__push__(Item(k))

    def __call__(self, *a: _T_Args, **kw: _T_Kwargs):
        return self.__push__(Call((a, FrozenDict(kw))))

    def __eval__(self, /, root: _T_Obj, start: int = None, stop: int = None):
        __tracebackhide__ = True
        val = root
        it = self.__ops__[start:stop]
        try:
            for t in it:
                val = t.__eval__(val)
        except EvaluationError:
            raise
        except Exception as e:
            raise EvaluationError.wrap(e) from e
        else:
            return val

    def __str__(self):
        return "<object>" + "".join(map(str, self.__expr__))

    def __iter__(self):
        return iter(self.__ops__)

    def __len__(self):
        return len(self.__expr__)

    # def __contains__(self, o):
    #     return o in self.__expr__

    def __reduce__(self):
        return self.__class__, self.__expr__


def look(expr: Lookup, /, root: _T_Obj, *, start: int = None, stop: int = None):
    return expr.__eval__(root, start=start, stop=stop)



import sys
import typing as t
from abc import ABC, abstractmethod
from collections import abc
from functools import wraps
from inspect import Parameter, Signature, iscoroutinefunction
from logging import getLogger
from types import FunctionType, GenericAlias

import attr
from typing_extensions import Self

from .graph.nodes import _T_Concrete, _T_Node

from ._common import lookups

from .graph import nodes
from ._common import Missing, FrozenDict, private_setattr, typed_signature
from ._functools import BoundParams
from .markers import Injectable, T_Injectable, T_Injected, is_injectable
from .markers import (
    GUARDED,
    PRIVATE,
    PROTECTED,
    PUBLIC,
    AccessModifier,
    Dep,
    DependencyMarker,
    Lookup,
    PureDep,
)

if sys.version_info < (3, 10):  # pragma: py-gt-39
    _UnionType = type(t.Union[t.Any, None])
else:  # pragma: py-lt-310
    from types import UnionType as _UnionType

_AnnotatedType = type(t.Annotated[t.Any, "ann"])


if t.TYPE_CHECKING:  # pragma: no cover
    from .containers import Container
    from .graph.core import DepKey, Graph


logger = getLogger(__name__)


_T = t.TypeVar("_T")
_T_Fn = t.TypeVar("_T_Fn", bound=abc.Callable, covariant=True)


def _fluent_decorator(fn=None, default=Missing, *, fluent: bool = False):
    def decorator(func: _T_Fn) -> _T_Fn:
        @wraps(func)
        def wrapper(self, v=default, /, *args, **kwds):
            nonlocal func, default, fluent
            if v is default:

                def decorator(val: _T) -> _T:
                    nonlocal func, v, args, kwds
                    rv = func(self, val, *args, **kwds)
                    return rv if fluent is True else val

                return decorator
            return func(self, v, *args, **kwds)

        wrapper._is_fluent_decorator = True

        return wrapper

    return decorator if fn is None else decorator(fn)


@DependencyMarker.register
@private_setattr(frozen="_frozen")
@attr.s(slots=True, frozen=True, cache_hash=True, cmp=True)
class Provider(t.Generic[_T_Concrete, _T_Node]):
    """The base class for all providers.

    Subclasses can implement the `resolve()` method to return the appropriate
    `Dependency` object for any given dependency.

    Attributes:
        concrete (Any): The object used to resolve
        container (Container): The Container where this provider is defined.
        access_modifier (t.Optional[AccessModifier]): The minimum access modifier required
            to access provider
        is_default (bool): Whether this provider is the default.
            A default provider only gets used if none other was provided to override it.
        is_final (bool): Whether this provider is final.
            A final provider will error if overridden by containers further
            down the provider resolution order.
        is_async (bool): Whether this provider is asyncnous

        filters (tuple[Callable]): Called to determine whether this provider can be resolved.

    """

    abstract: Injectable = None

    _frozen: bool = attr.ib(init=False, cmp=False, default=False)

    concrete: _T_Concrete = attr.ib(default=Missing)
    container: "Container" = attr.ib(kw_only=True, default=None)
    access_modifier: t.Optional[AccessModifier] = attr.ib(kw_only=True, default=None)

    is_default: bool = attr.ib(kw_only=True, default=False)
    is_final: bool = attr.ib(kw_only=True, default=False)

    is_async: bool = attr.ib(init=False, default=None)

    filters: tuple[abc.Callable[["Graph", Injectable], bool]] = attr.ib(
        kw_only=True, default=(), converter=tuple
    )

    _node_type: t.ClassVar[type[_T_Node]] = None
    _default_node_kwargs: t.ClassVar = {}

    __class_getitem__ = classmethod(GenericAlias)

    def default(self, is_default: bool = True) -> Self:
        """_Mark/Unmark_ this provider as the default. Updates the provider's
        `is_default` attribute.

        A default provider will be skipped if the dependency they provide has
        another provider. This means that a default provider will only get used
        if no other providers for the given dependency were defined in the current scope.

        Args:
            is_default (bool, optional): `True` to _mark_ or `False` to _unmark_.
                Defaults to `True`.
        Returns:
            self (Provider): this provider
        """
        self.__setattr(is_default=not not is_default)
        return self

    def final(self, is_final: bool = True) -> Self:
        """_Mark/Unmark_ this provider as final. Updates `is_final` attribute.

        A final provider will error if overridden by containers further down the
        provider resolution order. This does not apply to providers marked as
        default. Default providers will get skipped silently.

        Args:
            is_final (bool, optional): `True` to _mark_ or `False` to _unmark_.
                Defaults to `True`.
        Returns:
            self (Provider): this provider
        """
        self.__setattr(is_final=not not is_final)
        return self

    def private(self) -> Self:
        """Set the `access_modifier` for this provider to `AccessModifier.private`.

        A private provider is only avaliable to dependants declared in the same
        container.

        Returns:
            self (Provider): this provider
        """
        self.__setattr(access_modifier=PRIVATE)
        return self

    def guarded(self) -> Self:
        """Set the `access_modifier` for this provider to `AccessModifier.guarded`.

        A guarded provider is only visible to dependants declared in the
        provider's container and it's bases.

        Returns:
            self (Provider): this provider
        """
        self.__setattr(access_modifier=GUARDED)
        return self

    def protected(self) -> Self:
        """Set the `access_modifier` for this provider to `AccessModifier.protected`.

        A protected provider is only visible to dependants declared in containers
        within the inheritance heirachy of the provider's container. This includes
        both base and derived containers.

        Returns:
            self (Provider): this provider
        """
        self.__setattr(access_modifier=PROTECTED)
        return self

    def public(self) -> Self:
        """Set the `access_modifier` for this provider to `AccessModifier.public`.

        A public provider is visible to all dependants within it's scope.

        Returns:
            self (Provider): this provider
        """
        self.__setattr(access_modifier=PUBLIC)
        return self

    def _can_resolve(self, dep: "DepKey", scope: "Graph") -> bool:
        """Check whether this provider is avaliable to the given dep.

        Unlike `_resolve`, this method will be called on all matching dependencies.
        If available, the provider's `filters` will be applied and the result is
        returned.

        Returns `True` if no filters are available.

        Args:
            dep (DepKey): _description_
            scope (Scope): _description_

        Returns:
            bool: `True` if filters passed or `False` if otherwise
        """
        self._freeze()
        for fl in self.filters:
            if not fl(self, dep, scope):
                return False
        return True

    def _resolve(self, abstract: T_Injectable, scope: "Graph") -> _T_Node:
        """Resolves the given dependency.

        Args:
            abstract (T_Injectable): dependency to be resolved
            scope (Scope): Scope within which the dependency is getting resolved.

        Returns:
            node (Node):
        """
        return self._make_node(abstract, scope)

    def _setup(self, container: "Container", abstract: T_Injectable = None) -> Self:
        """Called when the provider is added to a container.

        Args:
            container (Container): the container
            abstract (T_Injectable): the bound dependency

        Raises:
            AttributeError: When another container was already set

        Returns:
            self (Provider): this provider
        """

        if not self.container is None:
            if not container is self.container:
                raise AttributeError(
                    f"container for `{self}` already set to `{self.container}`."
                )
        elif abstract and not self.abstract in (None, abstract):
            raise ValueError(
                f"invalid abstract type `{abstract}`. expected `{self.abstract}`"
            )
        else:
            self.__setattr(container=container)
        return self

    @t.overload
    def use(self) -> abc.Callable[[_T], _T]:
        ...  # pragma: no cover

    @t.overload
    def use(self, using: t.Any) -> Self:
        ...  # pragma: no cover

    @_fluent_decorator()
    def use(self, using: _T_Concrete) -> Self:
        """Set the provider's `concrete` attribute.

        The given value will depend on the type of provider

        Can be used as decorator

            @provider.use()
            def func():
                ...  # pragma: no cover

        Args:
            using (_T_Concrete): the object to provide. this depends on the type
                of provider

        Returns:
            self (Provider): this provider

        """
        self.__setattr(concrete=using)
        return self

    def when(self, *filters: abc.Callable, replace: bool = False) -> Self:
        """Add/replace the provider's `filters`.

        Filters are callables that determine whether this provider can provide a
        given dependency.

        Filters are called with 4 arguments:- `provider`- this provide,
        `abstract` - the dependency to be provided, `scope`- scope within which
        the dependency is getting resolved and `dependant` the `Container` from
        which the dependency was requested.

        Args:
            *filters (Callable): The filter callables.
            replace (bool, optional): Whether to replace the existing filters
                instead of adding. Defaults to False.

        Returns:
            self (Provider): this provider
        """
        if replace:
            self.__setattr(filters=tuple(dict.fromkeys(filters)))
        else:
            self.__setattr(filters=tuple(dict.fromkeys(self.filters + filters)))
        return self

    def _freeze(self):
        if not self._frozen:
            self._onfreeze()
            self.__setattr(_frozen=True)

    def _onfreeze(self):
        ...  # pragma: no cover

    def _node_kwargs(self, **kwds):
        return self._default_node_kwargs | kwds

    def _make_node(self, abstract: T_Injectable, scope: "Graph", **kwds):
        if cls := self._node_type:
            return cls(abstract, scope, self, **self._node_kwargs(**kwds))
        raise NotImplementedError(
            f"{self.__class__.__name__}._make_dependency()"
        )  # pragma: no cover


@attr.s(slots=True, frozen=True)
class Alias(Provider[T_Injectable, nodes._T_Node]):
    """Used to proxy another existing dependency. It resolves to the given `concrete`.

    For example. To use `_Ta` for dependency `_Tb`.

        container[_Tb] = Alias(_Ta)

    Args:
        concrete (Injectable): The dependency to be proxied
    """

    def _resolve(self, abstract: T_Injectable, scope: "Graph"):
        return scope[self.concrete]


@attr.s(slots=True, frozen=True)
class Value(Provider[T_Injected, nodes._T_ValueNode]):
    """Provides the given object as it is.

    Example:
        `T` will always resolve to `obj`.

            obj = object()
            container[T] = Value(obj)

    Args:
        concrete (T_Injected): The object to be provided
    """

    _node_type = nodes.Value

    def _node_kwargs(self, **kwds):
        kwds.setdefault("concrete", self.concrete)
        return super()._node_kwargs(**kwds)


@attr.s(slots=True, cmp=True, init=False, frozen=True)
class Factory(
    Provider[abc.Callable[..., T_Injected], nodes._T_FactoryNode],
    t.Generic[T_Injected, nodes._T_FactoryNode],
):
    """Resolves to the return value of the given factory. A factory can be a
    `type`, `function` or a `Callable` object.

    The factory is called every time a dependency for this provider is requested.

    Attributes:
        concrete (Union[type[T_Injected], abc.Callable[..., T_Injected]]): the factory to used
            to create the provided value.
        arguments (tuple[tuple. frozendict]): A tuple of positional and keyword
            arguments passed to the factory.

    Params:
        concrete (Union[type[T_Injected], abc.Callable[..., T_Injected]], optional):
            the factory. Can be a `type`, `function` or a `Callable` object.
        *args (Union[Dep, Any], optional): Positional arguments to pass to the factory.
        **kwargs (Union[Dep, Any], optional): Keyword arguments to pass to the factory.

    !!! note ""
        ## With Arguments

        Positional and/or keyword arguments to pass to the factory may be provided.

        ### Values Only

        ```python linenums="1" hl_lines="0"
        Factory(func, 'a', 32, obj, key='xyz')
        # will call: func('a', 32, obj, key='xyz')
        ```

        ### Values and Dependencies

        Arguments of type [`DependencyMarker`](../makers/#uzi.makers.DependencyMarker)
        will automatically be resolved and passed to the factory while calling it.

        e.g. using [`Dep`](../makers/#uzi.makers.Dep) and
        [`Lookup`](../makers/#uzi.makers.Lookup):
        ```python linenums="1" hl_lines="4 7"
        Factory(
            func,
            'a',
            uzi.makers.Dep(Foo),
            obj,
            key='xyz',
            bar=uzi.makers.Lookup(FooBar).bar
        )
        # will call: func('a', <inject: Foo>, obj, key='xyz', bar=<inject: FooBar>.bar)
        ```


    """

    arguments: tuple[tuple, FrozenDict] = attr.ib(default=((), FrozenDict()))
    # is_shared: t.ClassVar[bool] = False

    _signature: Signature = attr.ib(init=False, default=None)

    _blank_signature: t.ClassVar[Signature] = Signature()
    _arbitrary_signature: t.ClassVar[Signature] = Signature(
        [
            Parameter("__Parameter_var_positional", Parameter.VAR_POSITIONAL),
            Parameter("__Parameter_var_keyword", Parameter.VAR_KEYWORD),
        ]
    )

    _sync_node_type: t.ClassVar = nodes.Factory
    _asyns_node_type: t.ClassVar = nodes.AsyncFactory
    _await_params_sync_node_type: t.ClassVar = nodes.AwaitParamsFactory
    _await_params_async_node_type: t.ClassVar = nodes.AwaitParamsAsyncFactory

    def __init__(
        self, concrete: abc.Callable[[], T_Injectable] = None, /, *args, **kwargs
    ):
        self.__attrs_init__(concrete=concrete, arguments=(args, FrozenDict(kwargs)))

    def asynchronous(self, is_async: bool = True) -> Self:
        """_Mark/Unmark_ this provider as asynchronous. Updates `is_async`
        attribute.

        Normally, `coroutines` and factories with `async` dependencies automatically
        detected as asynchronous. This method provides the ability to change this
        default behaviour.

        Args:
            is_async (Union[bool, None], optional): `True` to _mark_, `False`
                to _unmark_ or `None` to revert to the default behaviour.
                Defaults to `True`.
        Returns:
            self (Provider): this provider
        """

        self.__setattr(is_async=is_async)
        return self

    def args(self, *args) -> Self:
        """Set the positional arguments to pass to the factory.

        Updates the `arguments` attribute.

        Args:
            *args (Union[Dep, Any], optional): Positional arguments to pass to
                the factory.

        Returns:
            self (Provider): this provider

        """
        arguments = self.arguments
        self.__setattr(arguments=(args, *arguments[1:]))
        return self

    def kwargs(self, **kwargs) -> Self:
        """Set the keyword arguments to pass to the factory.

        Updates the `arguments` attribute.

        Args:
            **kwargs (Union[Dep, Any], optional): Keyword arguments to pass to
                the factory.

        Returns:
            self (Provider): this provider
        """
        arguments = self.arguments
        self.__setattr(arguments=(*arguments[:1], FrozenDict(kwargs)))
        return self

    @t.overload
    def use(self) -> abc.Callable[[_T], _T]:
        pass

    @t.overload
    def use(self, using: abc.Callable, *args, **kwargs) -> Self:
        pass

    @_fluent_decorator()
    def use(self, concrete, *args, **kwargs):
        """Sets the provider's factory and arguments.

        Params:
            concrete (Union[type[T_Injected], abc.Callable[..., T_Injected]], optional):
                the factory. Can be a `type`, `function` or a `Callable` object.
            *args (Union[Dep, Any], optional): Positional arguments to pass to the factory.
            **kwargs (Union[Dep, Any], optional): Keyword arguments to pass to the factory.

        Returns:
            self (Factory): this provider
        """
        self.__setattr(concrete=concrete)
        args and self.args(*args)
        kwargs and self.kwargs(**kwargs)
        return self

    def signature(self, signature: Signature) -> Self:
        """Set a custom `Signature` for the factory.

        Args:
            signature (Signature): the signature

        Returns:
            self (Factory): this provider
        """
        self.__setattr(_signature=signature)
        return self

    def get_signature(self, dep: Injectable = None):
        sig = self._signature
        if sig is None:
            try:
                return typed_signature(self.concrete)
            except ValueError:
                return self._fallback_signature()
        return sig

    def _fallback_signature(self):
        return self._arbitrary_signature if self.arguments else self._blank_signature

    def _onfreeze(self):
        if None is self.is_async:
            self.__setattr(is_async=self._is_async_factory())

    def _is_async_factory(self) -> bool:
        return iscoroutinefunction(self.concrete)

    def _bind_params(
        self, scope: "Graph", abstract: Injectable, *, sig=None, arguments=()
    ):
        sig = sig or self.get_signature(abstract)
        args, kwargs = arguments or self.arguments
        return BoundParams.bind(sig, scope, self.container, args, kwargs)

    def _node_kwargs(self, **kwds):
        kwds.setdefault("concrete", self.concrete)
        return super()._node_kwargs(**kwds)

    def _make_node(self, abstract: T_Injectable, scope: "Graph", **kwds):
        params = self._bind_params(scope, abstract)
        if params.is_async:
            if self.is_async:
                cls = self._await_params_async_node_type
            else:
                cls = self._await_params_sync_node_type
        elif self.is_async:
            cls = self._asyns_node_type
        else:
            cls = self._sync_node_type

        return cls(
            abstract,
            scope,
            self,
            params=params,
            **self._node_kwargs(**kwds),
        )


@attr.s(slots=True, cmp=True, init=False)
class Singleton(Factory[T_Injected, nodes._T_SingletonNode]):
    """A `Singleton` provider is a `Factory` that returns same instance on every
    call.

    On the first request, the given factory will be called to create the instance
    which will be stored and returned on subsequent requests.

    Attributes:
        is_thread_safe (bool): Indicates whether to wrap the factory call with a
            `Lock` to prevent simultaneous instance create when injecting from
            multiple threads. Defaults to None
    """

    is_shared: t.ClassVar[bool] = True
    is_thread_safe: bool = attr.ib(init=False, default=None)

    _sync_node_type: t.ClassVar = nodes.Singleton
    _async_node_type: t.ClassVar = nodes.AsyncSingleton
    _await_params_sync_node_type: t.ClassVar = nodes.AwaitParamsSingleton
    _await_params_async_node_type: t.ClassVar = nodes.AwaitParamsAsyncSingleton

    def thread_safe(self, is_thread_safe: bool = True) -> Self:
        """_Mark/Unmark_ this provider as thread safe. Updates the `is_thread_safe`
        attribute.

        `is_thread_safe` indicates whether to wrap the factory call with a `Lock`
        to prevent simultaneous instance create when injecting from multiple threads.

        Args:
            is_thread_safe (bool, optional): `True` to _mark_ or `False` to
                _unmark_. Defaults to True.

        Returns:
            self (Provider): this provider
        """
        self.__setattr(is_thread_safe=is_thread_safe)
        return self

    def _node_kwargs(self, **kwds):
        kwds.setdefault("thread_safe", self.is_thread_safe)
        return super()._node_kwargs(**kwds)


@attr.s(slots=True, cmp=True, init=False)
class Resource(Singleton[T_Injected, nodes._T_ResourceNode]):
    """A `Resource` provider is a `Singleton` that has initialization and/or
    teardown.

    """

    is_async: bool = attr.ib(init=False, default=None)
    is_awaitable: bool = attr.ib(init=False, default=None)
    is_shared: t.ClassVar[bool] = True

    def awaitable(self, is_awaitable=True):
        self.__setattr(is_awaitable=is_awaitable)
        return self

    def _node_kwargs(self, **kwds):
        # kwds.setdefault('aw_enter', self.is_awaitable)
        return super()._node_kwargs(**kwds)


@attr.s(slots=True, init=False)
class Partial(Factory[T_Injected, nodes._T_PartialNode]):
    """A `Factory` provider that accepts extra arguments during resolution.

    Used internally to inject entry-point functions.
    """

    _sync_node_type: t.ClassVar = nodes.Partial
    _async_node_type: t.ClassVar = nodes.AsyncPartial
    _await_params_sync_node_type: t.ClassVar = nodes.AwaitParamsPartial
    _await_params_async_node_type: t.ClassVar = nodes.AwaitParamsAsyncPartial

    def _fallback_signature(self):
        return self._arbitrary_signature


@attr.s(slots=True, init=False)
class Callable(Partial[T_Injected, nodes._T_CallableNode]):
    """Similar to a `Factory` provider, a `Callable` provider resolves to a
    callable that wraps the factory.

    """

    _sync_node_type: t.ClassVar = nodes.Callable
    _async_node_type: t.ClassVar = nodes.AsyncCallable
    _await_params_sync_node_type: t.ClassVar = nodes.AwaitParamsCallable
    _await_params_async_node_type: t.ClassVar = nodes.AwaitParamsAsyncCallable


@attr.s(slots=True, frozen=True)
class LookupMarkerProvider(Factory[lookups.look, nodes._T_FactoryNode]):
    """Provider for resolving `uzi.Lookup` dependencies."""

    abstract = Lookup
    concrete = attr.ib(init=False, default=lookups.look)

    def _bind_params(self, scope: "Graph", marker: Lookup, *, sig=None, arguments=()):
        if not arguments:
            abstract = marker.__abstract__
            arguments = (lookups.Lookup(*marker),), FrozenDict(root=Dep(abstract))
        return super()._bind_params(scope, marker, sig=sig, arguments=arguments)


@attr.s(slots=True, frozen=True)
class UnionProvider(Provider[_T_Concrete]):
    """Provider for resolving `Union` dependencies."""

    abstract = t.get_origin(t.Union[t.Any, None])
    concrete = attr.ib(init=False, default=_UnionType)

    def get_all_args(self, abstract: Injectable):
        return t.get_args(abstract)

    def get_injectable_args(self, abstract: Injectable):
        return filter(is_injectable, self.get_all_args(abstract))

    def _resolve(self, abstract: T_Injectable, scope: "Graph"):
        for arg in self.get_injectable_args(abstract):
            if rv := scope[arg]:
                return rv


@attr.s(slots=True, frozen=True)
class AnnotationProvider(UnionProvider[_T_Concrete]):
    """Annotated types provider"""

    abstract = t.get_origin(t.Annotated[t.Any, object()])
    concrete = attr.ib(init=False, default=_AnnotatedType)

    def get_all_args(self, abstract: t.Annotated):
        logger.info(f"get_all_args --> {abstract}")
        logger.info(f" --> {abstract.__metadata__=}")

        for a in abstract.__metadata__[::-1]:
            if isinstance(a, DependencyMarker):
                yield a
                logger.info(f"yield --> {a}")
        yield abstract.__origin__


@attr.s(slots=True, frozen=True)
class DepMarkerProvider(Provider[_T_Concrete]):
    """Provider for resolving `zdi.Dep` dependencies."""

    abstract = Dep
    concrete = attr.ib(init=False, default=Dep)
    _node_type = nodes.Value

    def _resolve(self, marker: Dep, scope: "Graph") -> nodes.Node:
        dep = scope.make_key(marker.abstract, predicate=marker.predicate)
        if bind := scope[dep]:
            return bind
        elif marker.injects_default:
            return scope[marker.default]
        elif marker.has_default:
            return self._make_node(marker, scope, concrete=marker.default)


_T_Fn = t.TypeVar("_T_Fn", bound=abc.Callable)


def _provder_factory_method(cls: _T_Fn) -> _T_Fn:
    @wraps(cls)
    def wrapper(self: "ProviderRegistryMixin", abstract, *a, **kw):
        if not a:
            a = (abstract,)
        self[abstract] = pro = cls(*a, **kw)
        return pro

    return t.cast(cls, wrapper)


class ProviderRegistryMixin(ABC):
    """Implements a collection of helper methods for creating providers.

    Subclassed by `Container` to provide these methods
    """

    __slots__ = ()

    @abstractmethod
    def __setitem__(self, abstract: Injectable, provider: Provider):  # pragma: no cover
        ...  # pragma: no cover

    def provide(
        self, *providers: t.Union[Provider, type, t.TypeVar, abc.Callable], **kwds
    ) -> Self:
        for provider in providers:
            if isinstance(provider, tuple):
                abstract, provider = provider
            else:
                abstract, provider = provider, provider

            if isinstance(provider, Provider):
                if abstract == provider:
                    abstract = getattr(provider, "abstract", abstract)
                self[abstract] = provider
            elif isinstance(provider, (type, GenericAlias, FunctionType)):
                self.factory(abstract, provider, **kwds)
            elif abstract != provider:
                self.value(abstract, provider, **kwds)
            else:
                raise ValueError(
                    f"providers must be of type `Provider`, `type`, "
                    f"`FunctionType` not {provider.__class__.__name__}"
                )
        return self

    if t.TYPE_CHECKING:  # pragma: no cover

        def alias(self, abstract: Injectable, alias: t.Any, *a, **kw) -> Alias:
            ...  # pragma: no cover

        def value(self, abstract: Injectable, value: t.Any, *a, **kw) -> Value:
            ...  # pragma: no cover

        def callable(
            self, abstract: Injectable, factory: _T_Fn = ..., *a, **kw
        ) -> Callable:
            ...  # pragma: no cover

        def factory(
            self, abstract: Injectable, factory: _T_Fn = ..., *a, **kw
        ) -> Factory:
            ...  # pragma: no cover

        def resource(
            self, abstract: Injectable, factory: _T_Fn = ..., *a, **kw
        ) -> Resource:
            ...  # pragma: no cover

        def singleton(
            self, abstract: Injectable, factory: _T_Fn = ..., *a, **kw
        ) -> Singleton:
            ...  # pragma: no cover

    alias = _provder_factory_method(Alias)
    value = _provder_factory_method(Value)
    callable = _provder_factory_method(Callable)
    factory = _provder_factory_method(Factory)
    resource = _provder_factory_method(Resource)
    singleton = _provder_factory_method(Singleton)



from blinker import NamedSignal


on_container_create = NamedSignal(f"{__package__}.on_container_create")
on_container_init = NamedSignal(f"{__package__}.on_container_init")
on_provider_registered = NamedSignal(f"{__package__}.on_provider_registered")
on_injector_init = NamedSignal(f"{__package__}.on_injector_init")
on_scope_init = NamedSignal(f"{__package__}.on_scope_init")



from abc import ABC, abstractmethod
import logging
from threading import Lock
from typing_extensions import Self
import attr
import typing as t

from collections.abc import Callable

from .._common import Missing, private_setattr
from .._functools import (
    BoundParams,
    _PositionalArgs,
    _PositionalDeps,
    _KeywordDeps,
    FutureFactoryWrapper,
    FutureResourceWrapper,
    FutureCallableWrapper,
)

from ..markers import T_Injectable, T_Injected
from ..exceptions import InjectorLookupError

if t.TYPE_CHECKING:  # pragma: no cover
    from ..providers import Provider
    from ..injectors import Injector
    from ..containers import Container
    from . import Graph


_object_new = object.__new__


_T_Concrete = t.TypeVar("_T_Concrete")
"""Provided `concrete` `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
@private_setattr
class Node(ABC, t.Generic[_T_Concrete]):
    """A dependency graph node."""

    @abstractmethod
    def bind(self, injector: "Injector") -> t.Union[Callable[..., T_Injected], None]:
        raise NotImplementedError(
            f"{self.__class__.__name__}.bind()"
        )  # pragma: no cover

    abstract: T_Injectable = attr.ib()
    graph: "Graph" = attr.ib()
    provider: "Provider" = attr.ib(default=None, repr=lambda p: str(p and id(p)))

    concrete: _T_Concrete = attr.ib(kw_only=True, default=Missing, repr=True)

    is_async: bool = False
    dependencies = frozenset[Self]()

    _v_ident: tuple = attr.ib(init=False, repr=False)

    @_v_ident.default
    def _init_v_ident(self):
        return self.abstract, self.graph, self.container

    _ash: tuple = attr.ib(init=False, repr=False)

    @_ash.default
    def _init_ash(self):
        return hash(self._v_ident)

    @property
    def container(self):
        if pro := self.provider or self.graph:
            return pro.container

    def __eq__(self, o: Self) -> bool:
        if o.__class__ is self.__class__:
            return o._v_ident == self._v_ident
        elif isinstance(o, Node):
            return False
        return NotImplemented

    def __ne__(self, o: Self) -> bool:
        if o.__class__ is self.__class__:
            return o._v_ident != self._v_ident
        elif isinstance(o, Node):
            return True
        return NotImplemented

    def __hash__(self) -> int:
        return self._ash


_T_Node = t.TypeVar("_T_Node", bound=Node, covariant=True)
"""Node `TypeVar`"""


@private_setattr
class MissingNode:

    __slots__ = (
        "abstract",
        "graph",
    )

    graph: "Graph"
    container = None
    is_async: bool = False
    dependencies = frozenset[Self]()

    def __new__(
        cls: type[Self],
        abstract=None,
        graph: "Graph" = None,
        provider=None,
        concrete=None,
    ) -> Self:
        self = _object_new(cls)
        self.__setattr(abstract=abstract, graph=graph)
        return self

    def __bool__(self):
        return False

    def __reduce__(self):
        return self.__class__, (self.abstract, self.graph)

    def __eq__(self, o):
        return self.abstract == o

    def __ne__(self, o):
        return not self == o

    def __hash__(self) -> int:
        return hash(self.abstract)

    def bind(self, injector: "Injector"):
        raise InjectorLookupError(self.abstract, self.graph or injector.graph)


@attr.s(slots=True, frozen=True, cmp=False)
class SimpleNode(Node[_T_Concrete]):
    def bind(self, injector: "Injector"):
        return self.concrete(injector)


@attr.s(slots=True, frozen=True, cmp=False)
class Value(Node[T_Injected]):
    """Value node"""

    concrete: T_Injected = attr.ib(kw_only=True, default=None)
    is_async: t.Final = False

    def bind(self, injector: "Injector"):
        return self

    def __call__(self) -> T_Injected:
        return self.concrete


_T_ValueNode = t.TypeVar("_T_ValueNode", bound=Value, covariant=True)
"""Value node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class Factory(Node[T_Injected]):
    """Factory node"""

    concrete: T_Injected = attr.ib(kw_only=True)
    params: "BoundParams" = attr.ib(kw_only=True, default=BoundParams.make(()))
    thread_safe: bool = attr.ib(kw_only=True, default=False)

    @property
    def dependencies(self):
        return self.params.dependencies

    def bind(self: Self, injector: "Injector"):
        if self.params:
            args = self.resolve_args(injector)
            kwargs = self.resolve_kwargs(injector)
            vals = self.params.vals
            func = self.concrete

            def factory():
                nonlocal func, args, kwargs, vals
                return func(*args, **kwargs, **vals)

            return factory
        else:
            return self.concrete

    def resolve_args(self, injector: "Injector"):
        params = self.params
        if params.args:
            if params._pos_vals > 0 < params._pos_deps:
                return _PositionalArgs(
                    (p.value, None) if p.has_value else (None, injector[p.dependency])
                    for p in params.args
                )
            elif params._pos_deps > 0:
                return _PositionalDeps(injector[p.dependency] for p in params.args)
            else:
                return tuple(p.value for p in params.args)
        return ()

    def resolve_kwargs(self, injector: "Injector"):
        return _KeywordDeps((p.key, injector[p.dependency]) for p in self.params.kwds)


_T_FactoryNode = t.TypeVar("_T_FactoryNode", bound=Factory, covariant=True)
"""Factory node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class AsyncFactory(Factory[T_Injected]):
    """AsyncFactory node"""

    is_async: bool = True
    async_call: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsFactory(Factory[T_Injected]):
    """AwaitParamsFactory node"""

    is_async: bool = True
    async_call: bool = False

    def bind(self: Self, injector: "Injector"):
        args, aw_args = self.resolve_args(injector)
        kwargs, aw_kwargs = self.resolve_kwargs(injector)
        return FutureFactoryWrapper(
            self.concrete,
            self.params.vals,
            args=args,
            kwargs=kwargs,
            aw_args=aw_args,
            aw_kwargs=aw_kwargs,
            aw_call=self.async_call,
        )

    def resolve_kwargs(self, injector: "Injector"):
        if self.params.aw_kwds:
            deps = super().resolve_kwargs(injector)
            return deps, tuple((n, deps.pop(n)) for n in self.params.aw_kwds)
        else:
            return super().resolve_kwargs(injector), ()

    def resolve_args(self, injector: "Injector"):
        return super().resolve_args(injector), self.params.aw_args


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncFactory(AwaitParamsFactory[T_Injected]):
    """AwaitParamsAsyncFactory node"""

    async_call: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class Singleton(Factory[T_Injected]):
    """Singleton node"""

    # aw_enter: bool = attr.ib(kw_only=True, default=False)

    def factory(self, injector: "Injector"):
        if self.params:
            args = self.resolve_args(injector)
            kwargs = self.resolve_kwargs(injector)
            vals = self.params.vals
            func = self.concrete
            return lambda: func(*args, **kwargs, **vals)
        else:
            return self.concrete

    def bind(self, injector: "Injector"):
        func = self.factory(injector)
        value = Missing
        lock = Lock() if self.thread_safe else None

        def factory():
            nonlocal func, value
            if not value is Missing:
                return value

            lock and lock.acquire(blocking=True)
            try:
                if value is Missing:
                    value = func()
            finally:
                lock and lock.release()
            return value

        return factory


_T_SingletonNode = t.TypeVar("_T_SingletonNode", bound=Singleton, covariant=True)
"""Singleton node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class AsyncSingleton(Singleton[T_Injected]):
    """AsyncSingleton node"""

    is_async: bool = True
    async_call: bool = True

    def factory(self, injector: "Injector"):
        if params := self.params:
            args, kwargs = self.resolve_args(injector), self.resolve_kwargs(injector)
            return FutureFactoryWrapper(
                self.concrete,
                params.vals,
                args=args,
                kwargs=kwargs,
                aw_call=self.async_call,
            )
        else:
            return FutureFactoryWrapper(
                self.concrete, self.params.vals, aw_call=self.async_call
            )


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsSingleton(Singleton[T_Injected], AwaitParamsFactory[T_Injected]):
    """AwaitParamsSingleton node"""

    is_async: bool = True
    async_call: bool = False

    def factory(self, injector: "Injector"):
        (args, aw_args), (kwargs, aw_kwargs) = self.resolve_args(
            injector
        ), self.resolve_kwargs(injector)
        return FutureFactoryWrapper(
            self.concrete,
            self.params.vals,
            args=args,
            kwargs=kwargs,
            aw_args=aw_args,
            aw_kwargs=aw_kwargs,
            aw_call=self.async_call,
        )


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncSingleton(AwaitParamsSingleton[T_Injected]):
    """AwaitParamsAsyncSingleton node"""

    async_call: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class Resource(Singleton[T_Injected]):
    """Binds resources."""


_T_ResourceNode = t.TypeVar("_T_ResourceNode", bound=Resource, covariant=True)
"""Resource node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class Partial(Factory[T_Injected]):
    """Partial node"""

    def factory(self: Self, injector: "Injector"):
        args = self.resolve_args(injector)
        kwargs = self.resolve_kwargs(injector)
        vals = self.params.vals
        func = self.concrete

        def make(*a, **kw):
            nonlocal func, args, kwargs, vals
            return func(*args, *a, **(vals | kw), **kwargs.skip(kw))

        return make

    def bind(self: Self, injector: "Injector"):
        return self.factory(injector)


_T_PartialNode = t.TypeVar("_T_PartialNode", bound=Partial, covariant=True)
"""Partial node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class AsyncPartial(Partial[T_Injected]):
    """AsyncPartial node"""

    async_call: bool = True
    is_async: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsPartial(Partial[T_Injected], AwaitParamsFactory[T_Injected]):
    """AwaitParamsPartial node"""

    async_call: bool = False
    is_async: bool = True

    def factory(self: Self, injector: "Injector"):
        (args, aw_args), (kwargs, aw_kwargs) = self.resolve_args(
            injector
        ), self.resolve_kwargs(injector)
        return FutureCallableWrapper(
            self.concrete,
            self.params.vals,
            args=args,
            kwargs=kwargs,
            aw_args=aw_args,
            aw_kwargs=aw_kwargs,
            aw_call=self.async_call,
        )


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncPartial(AwaitParamsPartial[T_Injected]):
    """AwaitParamsAsyncPartial node"""

    async_call: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class Callable(Partial[T_Injected]):
    """Callable node"""

    def bind(self: Self, injector: "Injector"):
        func = self.factory(injector)
        return lambda: func


_T_CallableNode = t.TypeVar("_T_CallableNode", bound=Callable, covariant=True)
"""Callable node `TypeVar`"""


@attr.s(slots=True, frozen=True, cmp=False)
class AsyncCallable(Callable[T_Injected], AsyncPartial[T_Injected]):
    """AsyncCallable node"""

    is_async: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsCallable(Callable[T_Injected], AwaitParamsPartial[T_Injected]):
    """AwaitParamsCallable node"""

    is_async: bool = True


@attr.s(slots=True, frozen=True, cmp=False)
class AwaitParamsAsyncCallable(
    Callable[T_Injected], AwaitParamsAsyncPartial[T_Injected]
):
    """AwaitParamsAsyncCallable node"""

    async_call: bool = True



import typing as t
from collections import abc
from contextvars import ContextVar
from logging import getLogger

from typing_extensions import Self

from .. import Injectable
from .._common import FrozenDict, Missing, ReadonlyDict, private_setattr
from ..exceptions import FinalProviderOverrideError, ProError
from ..markers import (
    ProNoopPredicate,
    ProPredicate,
    _noop_pred,
    is_dependency_marker,
    is_injectable,
)
from .nodes import MissingNode, _T_Node

if t.TYPE_CHECKING:  # pragma: no cover
    from ..containers import Container
    from ..providers import Provider
    from ..scopes import Scope


logger = getLogger(__name__)

_T_Pro = tuple["Container"]
_T_BindKey = t.Union["DepKey", Injectable]

_object_new = object.__new__


class DepSrc(t.NamedTuple):
    graph: "Graph"
    container: "Container"
    predicate: ProPredicate = _noop_pred


@private_setattr
class DepKey:

    __slots__ = (
        "abstract",
        "src",
        "_ash",
    )

    abstract: Injectable
    src: DepSrc

    graph: "Graph" = None

    def __init_subclass__(cls, scope=None) -> None:
        cls.graph = scope
        return super().__init_subclass__()

    def __new__(
        cls: type[Self],
        abstract: Injectable,
        container: "Container" = None,
        predicate: ProPredicate = ProNoopPredicate(),
    ) -> Self:
        self, src = _object_new(cls), DepSrc(
            cls.graph, container, predicate or _noop_pred
        )
        self.__setattr(abstract=abstract, src=src, _ash=hash((abstract, src)))
        return self

    @property
    def container(self):
        return self.src.container

    @property
    def predicate(self):
        return self.src.predicate

    def replace(
        self,
        *,
        abstract: Injectable = None,
        container: "Container" = None,
        predicate: ProPredicate = None,
    ):
        return self.__class__(
            abstract or self.abstract,
            container or self.container,
            predicate or self.predicate,
        )

    def __eq__(self, o) -> bool:
        if isinstance(o, DepKey):
            return o.abstract == self.abstract and o.src == self.src
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, DepKey):
            return o.abstract != self.abstract or o.src != self.src
        return NotImplemented

    def __hash__(self) -> int:
        return self._ash


@private_setattr
class ProPaths(ReadonlyDict[DepSrc, _T_Pro]):
    __slots__ = (
        "graph",
        "pro",
    )

    graph: "Graph"
    pro: FrozenDict["Container", int]

    __contains = dict.__contains__
    __setdefault = dict.setdefault

    def __init__(self, graph: "Graph"):
        base = graph.parent.pros
        pro = {c: i for i, c in enumerate(graph.container.pro) if not c in base}
        if not pro:
            raise ProError(f"{graph.name}")
        self.__setattr(graph=graph, pro=FrozenDict(pro))

    def __contains__(self, x) -> bool:
        return x in self.pro or self.__contains(x)

    def __missing__(self, src: DepSrc):
        pro, graph = tuple(self.pro), self.graph
        src.graph.extends(graph)
        pro = src.predicate.pro_entries(pro, graph, src)
        return self.__setdefault(src, tuple(pro))


@private_setattr
class Graph(ReadonlyDict[_T_BindKey, _T_Node]):
    """An isolated dependency resolution `graph` for a given container.

    Assembles the dependency graphs of dependencies registered in their container.

    Attributes:
        container (Container): The container who's graph we are creating
        parent (Graph): The parent graph. Defaults to None

    Args:
        container (Container): The container who's graph we are creating
        parent (Graph, optional): The parent graph. Defaults to NullGraph

    """

    __slots__ = "container", "parent", "pros", "stack", "keyclass"

    container: "Container"
    parent: Self
    pros: ProPaths
    stack: "ResolutionStack"
    keyclass: type[DepKey]

    __contains = dict.__contains__
    __setdefault = dict.setdefault

    def __init__(self, container: "Container", parent: "Graph" = None):
        self.__setattr(
            container=container,
            parent=_null_graph if parent is None else parent,
            keyclass=type(f"BindKey", (DepKey,), {"graph": self}),
        )
        self.__setattr(
            pros=ProPaths(self),
            stack=ResolutionStack(container),
        )

    @property
    def level(self) -> int:
        return self.parent.level + 1

    @property
    def name(self):
        return self.container.name

    def parents(self):
        """Returns a generetor that iterates over the graph's ancestor starting
        from the current `parent` to the root graph.

        Yields:
            ancestor (Graph): an ancestor.
        """
        parent = self.parent
        while parent:
            yield parent
            parent = parent.parent

    def __bool__(self):
        return True

    def __contains__(self, o) -> bool:
        return self.__contains(o) or o in self.pros or o in self.parent

    def extends(self, graph: Self):
        return graph is self or self.parent.extends(graph)

    def make_key(
        self,
        abstract: Injectable,
        container: "Container" = None,
        predicate: ProPredicate = ProNoopPredicate(),
    ):
        if isinstance(abstract, DepKey):
            return abstract
        else:
            return self.keyclass(
                abstract, container or (self.stack.top.container), predicate
            )

    def find_provider(self, dep: DepKey):
        rv = [p for c in self.pros[dep.src] for p in c._resolve(dep, self)]
        if rv:
            if len(rv) > 1:
                rv.sort(key=lambda p: int(not not p.is_default))
                if final := next((p for p in rv if p.is_final), None):
                    if overrides := rv[: rv.index(final)]:
                        raise FinalProviderOverrideError(dep, final, overrides)
            return rv[0]

    def resolve(self, dep_: _T_BindKey, *, recursive: bool = True):
        if not (bind := self.get(dep_, Missing)) is Missing:
            if recursive or not bind or self is bind.graph:
                return bind
        elif dep_ != (dep := self.make_key(dep_)):
            bind = self.resolve(dep)
            if dep in self:
                bind = self.__setdefault(dep_, bind)
            if recursive or not bind or self is bind.graph:
                return bind
        elif is_injectable(dep.abstract):
            abstract = dep.abstract

            if prov := self.find_provider(dep):

                if prov.container and not prov.container is dep.container:
                    return self.__setdefault(
                        dep, self[self.make_key(abstract, prov.container)]
                    )

                with self.stack.push(prov, abstract):
                    if bind := prov._resolve(abstract, self):
                        return self.__setdefault(dep, bind)
            elif origin := t.get_origin(abstract):
                if is_dependency_marker(origin):
                    if prov := self.find_provider(
                        dep.replace(abstract=t.get_origin(abstract))
                    ):
                        with self.stack.push(prov, abstract):
                            if bind := prov._resolve(abstract, self):
                                return self.__setdefault(dep, bind)
                elif bind := self.resolve(
                    dep.replace(abstract=origin), recursive=False
                ):
                    return self.__setdefault(dep, bind)

            if recursive and ((bind := self.parent[dep]) or dep in self.parent):
                return self.__setdefault(dep, bind)
        else:
            raise TypeError(
                f"expected an `Injectable` not `{dep.abstract.__class__.__qualname__}`"
            )

    __missing__ = resolve

    def __eq__(self, o) -> bool:
        if isinstance(o, Graph):
            return o is self
        return NotImplemented

    def __ne__(self, o) -> bool:
        if isinstance(o, Graph):
            return not o is self
        return NotImplemented

    def __hash__(self):
        return id(self)


class NullGraph(Graph):
    """A 'noop' `Graph` used as the parent of root scopes.

    Attributes:
        container (frozendict):
        parent (None): The parent graph

    Params:
        None

    """

    __slots__ = ()
    parent = None
    container = FrozenDict()
    pros = FrozenDict()
    level = -1
    ident = ()
    _ash = hash(ident)

    name = "<null>"

    def __init__(self) -> None:
        ...  # pragma: no cover

    def extends(self, graph: Self):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __contains__(self, key):
        return False

    def __getitem__(self, key):
        if is_injectable(key):
            return MissingNode(key, self)
        elif isinstance(key, DepKey) and is_injectable(key.abstract):
            return MissingNode(key.abstract, self)
        else:
            raise TypeError(
                f"Graph keys must be `Injectable` not `{key.__class__.__qualname__}`"
            )

    def __eq__(self, o) -> bool:
        return o.__class__ is self.__class__

    def __ne__(self, o) -> bool:
        return not o.__class__ is self.__class__

    __hash__ = classmethod(hash)


_null_graph = NullGraph()


@private_setattr
class ResolutionStack(abc.Sequence):

    __slots__ = ("__var",)

    class StackItem(t.NamedTuple):
        container: "Container"
        abstract: Injectable = None
        provider: "Provider" = None

    __var: ContextVar[tuple[StackItem]]

    def __init__(self, default: "Container"):
        stack = (self.StackItem(default),)
        self.__var = ContextVar(
            f"{default.name}.{self.__class__.__name__}", default=stack
        )
        self.__var.set(stack)

    @property
    def top(self):
        return self[0]

    def push(self, provider: "Provider", abstract: Injectable = None):
        top = self.top
        new = self.StackItem(
            provider.container or top.container,
            abstract or provider.abstract or top.abstract,
            provider,
        )
        self.__var.set((new,) + self[:])
        return self

    def pop(self):
        var = self.__var
        stack = var.get()
        if len(stack) < 2:
            raise ValueError(f"too many calls to pop()")
        var.set(stack[1:])
        return stack[0]

    def index(self, val, start=0, stop=None):
        stack = self.__var.get()[start:stop:]

        if isinstance(val, tuple):
            return stack.index(val)
        else:
            for i, x in enumerate(stack):
                if val in x:
                    return i
        raise ValueError(val)

    def __reversed__(self):
        yield from reversed(self.__var.get())

    def __contains__(self, k):
        stack = self.__var.get()
        if isinstance(k, tuple):
            return k in stack
        else:
            return any(k in x for x in stack)

    def __getitem__(self, k):
        return self.__var.get()[k]

    def __bool__(self):
        return True

    def __len__(self):
        return len(self.__var.get())

    def __iter__(self):
        return iter(self.__var.get())

    def __enter__(self):
        return self

    def __exit__(self, *e):
        self.pop()

    def __copy__(self, *a):
        raise TypeError(f"cannot copy {self.__class__.__qualname__}")

    __deepcopy__ = __reduce__ = __copy__



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uzi', 'uzi._common', 'uzi.graph']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.4.0,<22.0.0',
 'blinker>=1.4,<2.0',
 'typing-extensions>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'uzi',
    'version': '0.3.2',
    'description': 'Dependency injection library',
    'long_description': '# Uzi\n\n\n[![PyPi version][pypi-image]][pypi-link]\n[![Supported Python versions][pyversions-image]][pyversions-link]\n[![Build status][ci-image]][ci-link]\n[![Coverage status][codecov-image]][codecov-link]\n\n\n`Uzi` is a [dependency injection](https://en.wikipedia.org/wiki/Dependency_injection) framework for Python.\n\n## Install\n\nInstall from [PyPi](https://pypi.org/project/uzi/)\n\n```\npip install uzi\n```\n\n## Features\n\n- Async support: `uzi` will `await` for you.\n- Lots of Providers to choose from. E.g.\n[Value](https://pyuzi.github.io/uzi/basic/providers/value.html), \n[Alias](https://pyuzi.github.io/uzi/basic/providers/alias.html).\n- Extensibility through `Container` inheritance.\n- Multi scope support.\n- Fast: minus the cost of an additional stack frame, `uzi` resolves dependencies \nnearly as efficiently as resolving them by hand.\n\n\n## Links\n\n- __[Documentation][docs-link]__\n- __[API Reference][api-docs-link]__\n- __[Installation][install-link]__\n- __[Get Started][why-link]__\n- __[Contributing][contributing-link]__\n\n\n\n## Production\n\nThis package is currently under active development and is not recommended for production use.\n\nWill be production ready from version `v1.0.0` onwards.\n\n\n\n[docs-link]: https://pyuzi.github.io/uzi/\n[api-docs-link]: https://pyuzi.github.io/uzi/api/\n[install-link]: https://pyuzi.github.io/uzi/install.html\n[why-link]: https://pyuzi.github.io/uzi/why.html\n[contributing-link]: https://pyuzi.github.io/uzi/0.5.x/contributing.html\n[pypi-image]: https://img.shields.io/pypi/v/uzi.svg?color=%233d85c6\n[pypi-link]: https://pypi.python.org/pypi/uzi\n[pyversions-image]: https://img.shields.io/pypi/pyversions/uzi.svg\n[pyversions-link]: https://pypi.python.org/pypi/uzi\n[ci-image]: https://github.com/pyuzi/uzi/actions/workflows/workflow.yaml/badge.svg?event=push&branch=master\n[ci-link]: https://github.com/pyuzi/uzi/actions?query=workflow%3ACI%2FCD+event%3Apush+branch%3Amaster\n[codecov-image]: https://codecov.io/gh/pyuzi/uzi/branch/master/graph/badge.svg\n[codecov-link]: https://codecov.io/gh/pyuzi/uzi\n\n\nSee this release on GitHub: [v0.3.2](https://github.com/pyuzi/uzi/releases/tag/0.3.2)\n',
    'author': 'David Kyalo',
    'author_email': 'davidmkyalo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pyuzi/uzi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import attr


class UziException(Exception):
    """Base class for all internal exceptions."""


@attr.s()
class FinalProviderOverrideError(TypeError, UziException):
    """Raised when a final provider has an override."""

    abstract: "Injectable" = attr.ib(default=None)
    final: "Provider" = attr.ib(default=None)
    overrides: tuple["Provider"] = attr.ib(default=(), converter=tuple)


class InvalidStateError(UziException):
    """The operation is not allowed in this state."""


@attr.s()
class InjectorLookupError(KeyError, UziException):
    """Raised by ~Injector` when a missing dependency is requested.

    Args:
        abstract (Injectable): the missing dependency
    """

    abstract: "Injectable" = attr.ib(default=None)
    scope: "Graph" = attr.ib(default=None)


class ProError(TypeError, UziException):
    """Raised when there is an issue with provider resolution order (`pro`)
    consistency
    """


from .providers import Provider
from .markers import Injectable
from .graph.core import Graph



from . import signals


from .containers import Container
from .providers import (
    AnnotationProvider,
    UnionProvider,
    DepMarkerProvider,
    LookupMarkerProvider,
)


@signals.on_container_create.connect
def _register_implicit_providers(sender: type[Container], container: Container):
    provs = (
        UnionProvider(),
        DepMarkerProvider(),
        AnnotationProvider(),
        LookupMarkerProvider(),
    )
    for prov in provs:
        container[prov.abstract] = prov
